from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from os import environ
from typing import (
    AsyncIterable,
    Awaitable,
    Coroutine,
    Deque,
    Iterable,
    Mapping,
    Sequence,
)


from twisted.internet.defer import Deferred, DeferredSemaphore
from twisted.internet.error import ProcessDone, ProcessTerminated
from twisted.internet.interfaces import IReactorProcess
from twisted.internet.protocol import ProcessProtocol
from twisted.python.failure import Failure
from twisted.python.procutils import which

from ._type import T, R


@dataclass
class ProcessResult:
    """
    The result of running a process to completion.
    """

    status: int
    output: bytes
    invocation: Invocation

    def check(self) -> None:
        """
        make sure that this process didn't exit with error
        """
        if self.status != 0:
            raise RuntimeError(
                f"process {self.invocation.executable} {self.invocation.argv} "
                f"exited with error {self.status}\n"
                f"{self.output.decode('utf-8', 'replace')}"
            )


@dataclass
class InvocationProcessProtocol(ProcessProtocol):
    def __init__(self, invocation: Invocation, quiet: bool) -> None:
        super().__init__()
        self.invocation = invocation
        self.d = Deferred[int]()
        self.quiet = quiet
        self.output = b""
        self.errors = b""

    def show(self, data: bytes) -> None:
        if not self.quiet:
            print(
                f"{self.invocation.executable} {' '.join(self.invocation.argv)}:",
                data.decode("utf-8", "replace"),
            )

    def outReceived(self, outData: bytes) -> None:
        self.output += outData
        self.show(outData)

    def errReceived(self, errData: bytes) -> None:
        self.errors += errData
        self.show(errData)

    def processEnded(self, reason: Failure) -> None:
        pd: ProcessDone | ProcessTerminated = reason.value
        self.d.callback(pd.exitCode)


@dataclass
class Invocation:
    """
    A full command-line to be invoked.
    """

    executable: str
    argv: Sequence[str]

    async def __call__(
        self,
        *,
        env: Mapping[str, str] = environ,
        quiet: bool = False,
        workingDirectory: str | None = None,
    ) -> ProcessResult:
        from twisted.internet import reactor

        ipp = InvocationProcessProtocol(self, quiet)
        IReactorProcess(reactor).spawnProcess(
            ipp,
            self.executable,
            [self.executable, *self.argv],
            environ,
            workingDirectory,
        )
        value = await ipp.d
        if value != 0:
            raise RuntimeError(
                f"{self.executable} {self.argv} exited with error {value}"
            )
        return ProcessResult(value, ipp.output, self)


@dataclass
class Command:
    """
    A command is a reference to a potential executable on $PATH that can be
    run.
    """

    name: str

    def __getitem__(self, argv: str | tuple[str, ...]) -> Invocation:
        """ """
        return Invocation(which(self.name)[0], argv)

    async def __call__(
        self,
        *args: str,
        env: Mapping[str, str] = environ,
        quiet: bool = False,
        workingDirectory: str | None = None,
    ) -> ProcessResult:
        """
        Immedately run.
        """
        return await self[args](env=env, quiet=quiet, workingDirectory=workingDirectory)


@dataclass
class SyntaxSugar:
    """
    Syntax sugar for running subprocesses.

    Use like::

        await c.ls()
        await c["docker-compose"]("--help")

    """

    def __getitem__(self, name) -> Command:
        """ """
        return Command(name)

    def __getattr__(self, name) -> Command:
        """ """
        return Command(name)


# from twisted.internet import reactor
c = SyntaxSugar()


async def parallel(
    work: Iterable[Coroutine[Deferred[T], T, R]], parallelism: int = 10
) -> AsyncIterable[R]:
    """
    Perform the given work with a limited level of parallelism.
    """
    sem = DeferredSemaphore(parallelism)
    values: Deque[R] = deque()

    async def saveAndRelease(coro: Awaitable[R]) -> None:
        try:
            values.append(await coro)
        finally:
            sem.release()

    async def drain() -> AsyncIterable[R]:
        await sem.acquire()
        while values:
            yield values.popleft()

    for w in work:
        async for each in drain():
            yield each
        Deferred.fromCoroutine(saveAndRelease(w))

    for x in range(parallelism):
        async for each in drain():
            yield each



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from enum import Enum, auto
from tempfile import NamedTemporaryFile
from typing import AsyncIterable, Iterable

from ._spawnutil import c, parallel
from twisted.internet.defer import Deferred
from twisted.python.filepath import FilePath
from twisted.python.procutils import which
from wheel_filename import ParsedWheelFilename, parse_wheel_filename


class KnownArchitecture(Enum):
    x86_64 = auto()
    arm64 = auto()
    universal2 = auto()
    purePython = auto()


@dataclass(frozen=True)
class PlatformSpecifics:
    """ """

    os: str
    major: int
    minor: int
    architecture: KnownArchitecture


def specifics(pwf: ParsedWheelFilename) -> Iterable[PlatformSpecifics]:
    """
    Extract platform specific information from the given wheel.
    """
    for tag in pwf.platform_tags:
        splitted = tag.split("_", 3)
        print("split", splitted)
        if len(splitted) != 4:
            continue
        os, major, minor, arch = splitted
        try:
            parsedArch = KnownArchitecture[arch]
        except ValueError:
            continue
        yield PlatformSpecifics(os, int(major), int(minor), parsedArch)


def wheelNameArchitecture(pwf: ParsedWheelFilename) -> KnownArchitecture:
    """
    Determine the architecture from a wheel.
    """
    if pwf.abi_tags == ["none"] and pwf.platform_tags == ["any"]:
        return KnownArchitecture.purePython
    allSpecifics = list(specifics(pwf))
    if len(allSpecifics) != 1:
        raise ValueError(
            f"don't know how to handle multi-tag wheels {pwf!r} {allSpecifics!r}"
        )
    return allSpecifics[0].architecture


@dataclass
class FusedPair:
    arm64: FilePath[str] | None = None
    x86_64: FilePath[str] | None = None
    universal2: FilePath[str] | None = None


async def findSingleArchitectureBinaries(
    paths: Iterable[FilePath[str]],
) -> AsyncIterable[FilePath[str]]:
    """
    Find any binaries under a given path that are single-architecture (i.e.
    those that will not run on an older Mac because they're fat binary).
    """
    checkedSoFar = 0

    async def checkOne(path: FilePath[str]) -> tuple[FilePath[str], bool]:
        """
        Check the given path for a single-architecture binary, returning True
        if it is one and False if not.
        """
        nonlocal checkedSoFar
        if path.islink():
            return path, False
        if not path.isfile():
            return path, False
        checkedSoFar += 1
        # universal binaries begin "Mach-O universal binary with 2 architectures"
        if (checkedSoFar % 1000) == 0:
            print("?", end="", flush=True)
        isSingle = (await c.file("-b", path.path, quiet=True)).output.startswith(
            b"Mach-O 64-bit bundle"
        )
        return path, isSingle

    async for eachPath, isSingleBinary in parallel(
        (
            checkOne(subpath)
            for path in paths
            for subpath in path.walk(
                # if we are in a virtualenv, but our system site packages has
                # single-architecture binaries installed, we should not be
                # concerned with those, as they're not on our import path -
                # even though they *do* live below an entry on sys.path (the
                # stdlib).
                lambda sub: sub.basename()
                != "site-packages"
            )
        ),
        16,
    ):
        if isSingleBinary:
            yield eachPath


def determineNeedsFusing(
    downloadDir: str, fusedDir: str
) -> Iterable[tuple[tuple[str, str], FusedPair]]:
    needsFusing: defaultdict[tuple[str, str], FusedPair] = defaultdict(FusedPair)

    for child in FilePath(downloadDir).children():
        # every wheel in this list should either be architecture-independent,
        # universal2, *or* have *both* arm64 and x86_64 versions.
        pwf = parse_wheel_filename(child.basename())
        arch = wheelNameArchitecture(pwf)
        fusedPath = FilePath(fusedDir).child(child.basename())
        if arch == KnownArchitecture.purePython:
            child.moveTo(fusedPath)
            continue
        # OK we need to fuse a wheel
        fusor = needsFusing[(pwf.project, pwf.version)]
        if arch == KnownArchitecture.x86_64:
            fusor.x86_64 = child
        if arch == KnownArchitecture.arm64:
            fusor.arm64 = child
        if arch == KnownArchitecture.universal2:
            child.moveTo(fusedPath)
            fusor.universal2 = fusedPath
    return needsFusing.items()


async def fuseOne(
    tmpDir: str, fusedDir: str, name: str, version: str, fusor: FusedPair
) -> None:
    if fusor.universal2 is not None:
        print(f"{name} has universal2; skipping")
        return

    left = fusor.arm64
    if left is None:
        raise RuntimeError(f"no arm64 architecture for {name}")
    right = fusor.x86_64
    if right is None:
        raise RuntimeError(f"no x86_64 architecture for {name}")
    await c["delocate-merge"](
        "--verbose", f"--wheel-dir={fusedDir}", left.path, right.path
    )


async def fixArchitectures() -> None:
    """
    Ensure that all wheels installed in the current virtual environment are
    universal2, not x86_64 or arm64.

    This probably only works on an arm64 (i.e., Apple Silicon) machine since it
    requires the ability to run C{pip} under both architectures.
    """
    downloadDir = ".wheels/downloaded"
    tmpDir = ".wheels/tmp"
    fusedDir = ".wheels/fused"

    output = (await c.pip("freeze")).output.decode("utf-8")
    with NamedTemporaryFile(delete=False) as f:
        for line in output.split("\n"):
            if (":" not in line) and ("/" not in line) and (not line.startswith("-e")):
                f.write((line + "\n").encode("utf-8"))

    await c.mkdir("-p", downloadDir, fusedDir, tmpDir)
    for arch in ["arm64", "x86_64"]:
        await c.arch(
            f"-{arch}",
            which("pip")[0],
            "wheel",
            "-r",
            f.name,
            "-w",
            downloadDir,
        )

    async for each in parallel(
        fuseOne(tmpDir, fusedDir, name, version, fusor)
        for ((name, version), fusor) in determineNeedsFusing(downloadDir, fusedDir)
    ):
        pass

    await c.pip(
        "install",
        "--no-index",
        "--find-links",
        fusedDir,
        "--force",
        "--requirement",
        f.name,
    )


start = Deferred.fromCoroutine


async def validateArchitectures(
    paths: Iterable[FilePath[str]], report: bool = True
) -> bool:
    """
    Ensure that there are no problematic single-architecture binaries in a
    given directory.
    """
    success = True
    async for eachBinary in findSingleArchitectureBinaries(paths):
        if (
            # exclude py2app prebuilt executable stubs
            (
                eachBinary.basename() in {"main-x86_64", "main-arm64"}
                and eachBinary.parent().basename() == "prebuilt"
            )
            or
            # exclude debugpy attach stubs
            (
                eachBinary.basename() == "attach_x86_64.dylib"
                and eachBinary.parent().basename() == "pydevd_attach_to_process"
            )
            or
            # exclude delocate's own tests
            (
                eachBinary.parent().basename() == "data"
                and eachBinary.parent().parent().basename() == "tests"
                and eachBinary.parent().parent().parent().basename() == "delocate"
            )
        ):
            continue
        if report:
            print()
            print(f"single-architecture binary: {eachBinary.path}")
        success = False
    return success



from functools import wraps
from getpass import getpass
from json import load
from os import environ
from os.path import expanduser
from typing import (
    Any,
    Awaitable,
    Callable,
    Concatenate,
    Coroutine,
    Generator,
    ParamSpec,
    TypeVar,
)

import click

from ._build import AppBuilder, whichSetup
from ._spawnutil import c
from twisted.internet.defer import Deferred
from twisted.internet.task import react
from twisted.python.failure import Failure


P = ParamSpec("P")
R = TypeVar("R")


async def configuredBuilder() -> AppBuilder:
    """
    Make an AppBuilder out of the local configuration.
    """
    with open(expanduser("~/.encrust.json")) as f:
        obj = load(f)
    lines = await c.python(whichSetup(), "--name", "--version")
    name, version = lines.output.decode("utf-8").strip().split("\n")
    name = environ.get("ENCRUST_APP_NAME", name)
    version = environ.get("ENCRUST_APP_VERSION", version)
    return AppBuilder(
        name=name,
        version=version,
        identityHash=obj["identity"],
        notarizeProfile=obj["profile"],
        appleID=obj["appleID"],
        teamID=obj["teamID"],
    )


def reactorized(
    c: Callable[
        Concatenate[Any, P],
        Coroutine[Deferred[object], Any, object]
        | Generator[Deferred[object], Any, object],
    ]
) -> Callable[P, None]:
    """
    Wrap an async twisted function for click.
    """

    @wraps(c)
    def forclick(*a, **kw) -> None:
        def r(reactor: Any) -> Deferred[object]:
            async def ar():
                try:
                    await c(reactor, *a, **kw)
                except Exception:
                    print(Failure().getTraceback())

            return Deferred.fromCoroutine(ar())

        react(r, [])

    return forclick


@click.group()
def main() -> None:
    """
    Utility for building, signing, and notarizing macOS applications.
    """


@main.command()
@reactorized
async def fatten(reactor: Any) -> None:
    """
    Ensure that all locally installed shared objects are fat binaries (i.e.
    universal2 wheels).
    """
    builder = await configuredBuilder()
    await builder.fattenEnvironment()


@main.command()
@reactorized
async def build(reactor: Any) -> None:
    """
    Build the application.
    """
    builder = await configuredBuilder()
    await builder.build()
    await builder.signApp()


@main.command()
@reactorized
async def release(reactor: Any) -> None:
    """
    Build the application.
    """
    builder = await configuredBuilder()
    await builder.release()


@main.command()
@reactorized
async def auth(reactor: Any) -> None:
    """
    Authenticate to the notarization service with an app-specific password from
    https://appleid.apple.com/account/manage
    """
    builder = await configuredBuilder()
    newpw = getpass(
        f"Paste App-Specific Password for {builder.appleID} and hit enter: "
    )
    await builder.authenticateForSigning(newpw)
    print("Authenticated!")


@main.command()
@reactorized
async def configure(reactor: Any) -> None:
    """
    Configure this tool.
    """
    print(
        """
    TODO: this tool should walk you through configuration.

    For now:
    0. First, set up a Python project built using `py2app`.
        a. make a virtualenv
        b. `pip install` your dependencies
        c. `pip install encrust`

    1. enroll in the Apple Developer program at https://developer.apple.com/account
    2. download Xcode.app from https://apps.apple.com/us/app/xcode/id497799835?mt=12
    3. launch Xcode,
        a. open Preferences -> Accounts
        b. hit '+' to log in to the Apple ID you enrolled in
           the developer program with
        c. click "manage certificates"
        d. click "+"
        e. click "Developer ID Application"
    4. run `security find-identity -v -p codesigning`
    5. look for a "Developer ID Application" certificate in the list
    6. edit ~/.encrust.json to contain an object like this:

        {
            "identity": /* the big hex ID from find-identity output */,
            "teamID": /* the thing in parentheses in find-identity output */,
            "appleID": /* the email address associated with your apple developer account */,
            "profile": /* an arbitrary string you've selected */
        }
    7. go to https://appleid.apple.com/account/manage and log in
    8. click "App-Specific Passwords"
    9. run `encrust auth` and paste the app password before closing the window
    10. run `encrust release`
    11. upload dist/<YourApp>-<YourVersion>.release.app.zip somewhere on the web.
    """
    )



from __future__ import annotations


from twisted.python.filepath import FilePath

from ._spawnutil import c


async def createZipFile(zipFile: FilePath, directoryToZip: FilePath) -> None:
    zipPath = zipFile.asTextMode().path
    dirPath = directoryToZip.asTextMode()
    await c.ditto(
        "-c",
        "-k",
        "--sequesterRsrc",
        "--keepParent",
        dirPath.basename(),
        zipPath,
        workingDirectory=dirPath.dirname(),
    )



from __future__ import annotations

from typing import Iterable

from twisted.python.filepath import FilePath

from ._spawnutil import c


async def signOneFile(
    fileToSign: FilePath[str],
    codesigningIdentity: str,
    entitlements: FilePath[str],
) -> None:
    """
    Code sign a single file.
    """
    fileStr = fileToSign.path
    entitlementsStr = entitlements.path
    print("✓", end="", flush=True)
    await c.codesign(
        "--sign",
        codesigningIdentity,
        "--entitlements",
        entitlementsStr,
        "--force",
        "--options",
        "runtime",
        fileStr,
    )


def signablePathsIn(topPath: FilePath[str]) -> Iterable[FilePath[str]]:
    """
    What files need to be individually code-signed within a given bundle?
    """
    for p in topPath.walk():
        ext = p.splitext()[-1]
        if ext in {".so", ".dylib", ".framework", ".a", ".app"}:
            yield p
        elif p.basename() == 'python' and p.parent().basename() == "MacOS":
            yield p


async def notarize(
    *,
    archivePath: FilePath[str],
    applicationPath: FilePath[str],
    appleID: str,
    teamID: str,
    notarizeProfile: str,
) -> None:
    """
    Submit the signed bundle for notarization, wait for success, then notarize
    it.
    """
    await c.xcrun(
        "notarytool",
        "submit",
        archivePath.path,
        f"--apple-id={appleID}",
        f"--team-id={teamID}",
        f"--keychain-profile={notarizeProfile}",
        f"--wait",
    )
    await c.xcrun("stapler", "staple", applicationPath.path)



"""
Future work:

- integrate cocoapods
"""
from __future__ import annotations
import sys

from dataclasses import dataclass

from twisted.python.filepath import FilePath
from twisted.python.modules import getModule

from ._signing import notarize, signOneFile, signablePathsIn
from ._architecture import validateArchitectures, fixArchitectures
from ._spawnutil import c, parallel
from ._zip import createZipFile


def whichSetup() -> str:
    cwd = FilePath(".")
    for possibility in ['py2app_setup.py', 'setup.py']:
        if cwd.child(possibility).exists():
            return possibility
    raise Exception("no setup.py found")

@dataclass
class AppBuilder:
    """
    A builder for a particular application
    """

    name: str
    version: str
    notarizeProfile: str
    appleID: str
    teamID: str
    identityHash: str
    entitlementsPath: FilePath[str] = getModule(__name__).filePath.sibling(
        "required-python-entitlements.plist"
    )

    async def release(self) -> None:
        """
        Execute the release end to end; build, sign, archive, notarize, staple.
        """
        await self.fattenEnvironment()
        await self.build()
        archOK = await validateArchitectures([self.originalAppPath()], True)
        if not archOK:
            raise RuntimeError()
        await self.signApp()
        await self.notarizeApp()

    async def fattenEnvironment(self) -> None:
        """
        Ensure the current virtualenv has all universal2 "fat" binaries.
        """
        pathEntries = [FilePath(each) for each in sys.path if each]
        needsFattening = not await validateArchitectures(pathEntries)
        if not needsFattening:
            print("already ok")
            return
        await fixArchitectures()
        stillNeedsFattening = not await validateArchitectures(pathEntries, True)
        if stillNeedsFattening:
            raise RuntimeError(
                "single-architecture binaries still exist after fattening: {stillNeedsFattening}"
            )
        print("all relevant binaries now universal2")

    def archivePath(self, variant: str) -> FilePath[str]:
        """
        The path where we should archive our zip file.
        """
        return FilePath("dist").child(f"{self.name}-{self.version}.{variant}.app.zip")

    async def archiveApp(self, variant: str) -> FilePath[str]:
        """ """
        archivedAt = self.archivePath(variant)
        await createZipFile(archivedAt, self.originalAppPath())
        return archivedAt

    async def build(self) -> None:
        """
        Just run py2app.
        """
        await c.python(whichSetup(), "py2app")

    async def authenticateForSigning(self, password: str) -> None:
        """
        Prompt the user to authenticate for code-signing and notarization.
        """
        await c.xcrun(
            "notarytool",
            "store-credentials",
            self.notarizeProfile,
            "--apple-id",
            self.appleID,
            "--team-id",
            self.teamID,
            "--password",
            password,
        )

    def originalAppPath(self) -> FilePath[str]:
        """
        A L{FilePath} pointing at the application (prior to notarization).
        """
        return FilePath("./dist").child(self.name + ".app")

    async def signApp(self) -> None:
        """
        Find all binary files which need to be signed within the bundle and run
        C{codesign} to sign them.
        """
        top = self.originalAppPath()
        async for signResult in parallel(
            (
                signOneFile(p, self.identityHash, self.entitlementsPath)
                for p in signablePathsIn(top)
            )
        ):
            pass
        await signOneFile(top, self.identityHash, self.entitlementsPath)

    async def notarizeApp(self) -> None:
        """
        Submit the built application to Apple for notarization and wait until we
        have seen a response.
        """
        preReleasePath = await self.archiveApp("for-notarizing")
        await notarize(
            appleID=self.appleID,
            teamID=self.teamID,
            archivePath=preReleasePath,
            applicationPath=self.originalAppPath(),
            notarizeProfile=self.notarizeProfile,
        )
        await self.archiveApp("release")
        preReleasePath.remove()
